/************************************************************************/
/*                                                                      */
/* Copyright (C) 1997 Arve Kylling                                      */
/*                                                                      */
/* This program is free software; you can redistribute it and/or modify */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation; either version 1, or (at your option)  */
/* any later version.                                                   */
/*                                                                      */
/* This program is distributed in the hope that it will be useful,      */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY of FITNESS FOR A PARTICULAR PURPOSE. See the         */
/* GNU General Public License for more details.                         */
/*                                                                      */
/* To obtain a copy of the GNU General Public License write to the      */
/* Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,   */
/* USA.                                                                 */
/*                                                                      */
/************************************************************************/

#include "uvspec.h"

/**************************************************************/
/* The main uvspec function.                                  */
/**************************************************************/

int uvspec (input_struct input, output_struct *output)
{
  int status = 0;

  /**** Check the model input data ****/
  status = uvspec_check(input);
  if (status!=0) {
    fprintf (stderr, "%d error(s) in uvspec input-file, aborting\n", abs(status));
    return status;
  }


  /**** Set wavelength grid for the transmittance calculation ****/
  status = setup_wlgrid (input, output);
  if (status!=0) {
    fprintf (stderr, "Error %d setting up wavelength grid\n", status);
    return status;
  } 


  /**** Cloudless sky atmosphere ****/
  status = setup_atmosphere (input, output);
  if (status!=0) {
    fprintf (stderr, "Error %d setting up atmosphere\n", status);
    return status;
  }
  

  /**** Setup temperature at computing levels ****/
  status = setup_temperature (input, output);
  if (status!=0) {
    fprintf (stderr, "Error %d setting up temperature\n", status);
    return status;
  }
  

  /****  Read extraterrestrial spectrum and correct for Earth-Sun distance ****/
  status = setup_extraterrestrial (input, output);
  if (status!=0) {
    fprintf (stderr, "Error %d setting up extraterrestrial irradiance\n", status);
    return status;
  }


  /**** Read cross section files, like O3, BrO, OclO, NO2, etc. ****/
  status = setup_crs (input, output);
  if (status!=0) {
    fprintf (stderr, "Error %d setting up absorption cross sections\n", status);
    return status;
  }


  /**** Calculate Rayleigh scattering cross section ****/
  status = setup_rayleigh (input, output);
  if (status!=0) {
    fprintf (stderr, "Error %d setting up Rayleigh cross sections\n", status);
    return status;
  }


  /**** Setup wavelength dependent albedo ****/
  status = setup_albedo (input, output);
  if (status!=0) {
    fprintf (stderr, "Error %d setting up wavelength dependent albedo\n", status);
    return status;
  }
  

  /**** Setup sza and phi0 ****/
  status = setup_sza (input, output);
  if (status!=0) {
    fprintf (stderr, "Error %d setting up sza and phi0\n", status);
    return status;
  }
  

  /**** Ice cloud stuff ****/
  status = setup_ice (input, output);
  if (status!=0) {
    fprintf (stderr, "Error %d setting up ice cloud properties\n", status);
    return status;
  }


  /**** Water cloud stuff ****/
  status = setup_wcloud (input, output);
  if (status!=0) {
    fprintf (stderr, "Error %d setting up water cloud properties\n", status);
    return status;
  }


  /**** Independent pixel stuff ****/
  status = setup_ipa (input, output);
  if (status!=0) {
    fprintf (stderr, "Error %d setting up IPA properties\n", status);
    return status;
  }
  

  /****   Aerosol stuff   ****/
  status = setup_aerosol (input, output);
  if (status!=0) {
    fprintf (stderr, "Error %d setting up aerosol properties\n", status);
    return status;
  }


  /****     Solve RTE     ****/
  status = solve_rte (input, output);
  if (status!=0) {
    fprintf (stderr, "Error %d solving RTE\n", status);
    return status;
  }


  /**** Interpolate transmittance to high resolution wavelength grid ****/
  status = interpolate_transmittance (input, output);
  if (status!=0) {
    fprintf (stderr, "Error %d interpolating model output to hires wavelength grid\n", status);
    return status;
  }


  /**** Multiply with extraterrestrial irradiance ****/
  status = multiply_extraterrestrial (input, output);
  if (status!=0) {
    fprintf (stderr, "Error %d multiplying with extraterrestrial irradiance\n", status);
    return status;
  }


  /**** Convolve with slit function ****/
  if (input.convolve) {
    status = convolve (input, output);
    if (status!=0) {
      fprintf (stderr, "Error %d convolving hires spectrum with slit function\n", status);
      return status;
    }
  }


  /**** Finally interpolate to output grid ****/
  if (input.spline) {
    status = spline_interpolate (input, output);
    if (status!=0) {
      fprintf (stderr, "Error %d interpolating data spectrum to output grid\n", status);
      return status;
    }
  }

  
  /**** Wavelength-integrate 3D fields ****/
  if (input.rte.mc.passback3D && input.ck_scheme!=CK_NONE) {
    status = integrate3D (input, output);
    if (status!=0) {
      fprintf (stderr, "Error %d integrating 3D output over wavelength\n", status);
      return status;
    }
  }


  return 0;
}



/**************************************************************/
/* Check the model input for consistency.                     */
/**************************************************************/

int uvspec_check (input_struct Input)
{
  int i=0, status=0;
  int s1=0, s2=0;
  FILE *f=NULL;

  /* ***FIXME*** Still need to check (also in uvspec function):
     aerosol_haze
     aerosol_scale_ssa
     aerosol_set_gg
     aerosol_set_ssa
     aerosol_season
     aerosol_visibility
     aerosol_vulcan
     angstrom
     atmosphere_file
     data_files_path
     nstr
     rte_solver
     solar_file
     sza
     sza_file
     umu
     wc_file
     wc_set_gg
     wc_set_ssa
     wc_layer
     wvn
   */


  /* Albedo checks */
  
  if ( Input.atm.albedo > 1.0 ) {
    fprintf (stderr, "*** Surface albedo > 1.0: %f\n", Input.atm.albedo);
    status--;
  }
  else if ( Input.atm.albedo < 0.0 && strlen(Input.filename[FN_ALBEDO])==0 ) {
    fprintf (stderr, "*** Surface albedo < 0.0: %f\n", Input.atm.albedo);
    status--;
  }


  if ((strlen(Input.filename[FN_RPV])>0 || 
	      Input.rpv.rho0!=0 || Input.rpv.k!=0 || Input.rpv.theta!=0) && 
      (Input.rte.solver!=MONTECARLO && Input.rte.solver!=DISORT2)) {
    fprintf (stderr, "*** Error, RPV BDRF is only implemented with MYSTIC and DISORT2!\n");
    status--;
  }


  /* Day of year checks */

  if ( Input.day_of_year > 366 ) {
    fprintf (stderr, "*** Day of year > 366: %d\n", Input.day_of_year);
    status--;
  }
  else if ( Input.day_of_year < 0 && Input.day_of_year != -999 ) {
    fprintf (stderr, "*** Day of year < 0: %d\n", Input.day_of_year);
    status--;
  }

  /* Ozone column checks */

  if ( Input.ozone_column < 0.0 && Input.ozone_column != -999 ) {
    fprintf (stderr, "*** Ozone_column < 0.0: %f\n", Input.ozone_column);
    status--;
  }


  /* Aerosol checks */

  /* if any of the following four variables is set, all four must be set; */
  /* otherwise the information is incomplete!                             */
  if (Input.aer.seasn>0 || Input.aer.vulcan>0 || Input.aer.haze>0 || Input.aer.visib>0) {

    if (Input.aer.seasn<=0) {
      fprintf (stderr, "*** Aerosol information incomplete; specify aerosol_season or aerosol_default!\n");
      status--;
    }

    if (Input.aer.vulcan<=0) {
      fprintf (stderr, "*** Aerosol information incomplete; specify aerosol_vulcan or aerosol_default!\n");
      status--;
    }

    if (Input.aer.haze<=0) {
      fprintf (stderr, "*** Aerosol information incomplete; specify aerosol_haze or aerosol_default!\n");
      status--;
    }

    if (Input.aer.visib<=0) {
      fprintf (stderr, "*** Aerosol information incomplete; specify aerosol_visibility or aerosol_default!\n");
      status--;
    }
  }
  

  s1=strlen(Input.aer.filename[FN_AER_SIZ]);
  s2=strlen(Input.aer.filename[FN_AER_REF]);

  if (s1>0 && (s2==0 && Input.aer.re==0)) {
    fprintf (stderr, "*** Aerosol information incomplete; specify aerosol_refrac_index or aerosol_refrac_file\n");
    fprintf (stderr, "***                                 or omit the size distribution\n");
    status--;
  }

  if (s1==0 && (s2>0 || Input.aer.re>0)) {
    fprintf (stderr, "*** Aerosol information incomplete; specify aerosol_sizedist_file\n");
    fprintf (stderr, "***                                 or omit the refractive index!\n");
    status--;
  }

  
  if (strlen(Input.filename[FN_WCLOUD])>0 && strlen(Input.filename[FN_WCLOUD_FILES])>0 ) {
    fprintf (stderr, "*** It doesn't make sense to specify both wc_file and wc_files;\n");
    fprintf (stderr, "*** please remove either one and retry!\n");
    status--;
  }

  if (strlen(Input.filename[FN_WCLOUD_PROP])>0 && strlen(Input.filename[FN_WCLOUD])==0 ) {
    fprintf (stderr, "*** wc_properties only makes sense together with 'wc_file'!\n");
    status--;
  }

  if (strlen(Input.filename[FN_ICLOUD_PROP])>0 && strlen(Input.filename[FN_ICLOUD])==0 ) {
    fprintf (stderr, "*** ic_properties only makes sense together with 'ic_file'!\n");
    status--;
  }


  if (strlen(Input.filename[FN_WCLOUD_PROP])>0 && !Input.wc.layer) {
    fprintf (stderr, "*** wc_properties requires optical properties defined per layer.\n");
    fprintf (stderr, "*** Please specify wc_layer and retry, if that is what you want!\n");
    status--;
  }


  if (strlen(Input.filename[FN_ICLOUD_PROP])>0 && !Input.ic.layer) {
    fprintf (stderr, "*** ic_properties requires optical properties defined per layer.\n");
    fprintf (stderr, "*** Please specify ic_layer and retry, if that is what you want!\n");
    status--;
  }

  /* RTE checks */

  switch (Input.rte.solver) {
  case SDISORT:

    if (strlen(Input.filename[FN_WCLOUD])>0 ) {
      fprintf (stderr, "*** sdisort is not suited for calculations involving clouds\n");
      status--;
    }

    if (Input.source==SRC_THERMAL) {
      fprintf (stderr, "*** sdisort does not include thermal emission\n");
      status--;
    }

    break;

  case TWOSTR:

    if (Input.rte.nprndis>0 )
      for (i=0;i<Input.rte.nprndis;i++)
	if ( !(Input.rte.prndis[i] == 1 || Input.rte.prndis[i] == 2) ) { 
	  fprintf (stderr, "*** twostr can only handle prndis set to 1 and/or 2\n");
	  status--;
	}

    break;

  case POLRADTRAN:
    break;

  case MONTECARLO:
    break;

  default:
    break;
  }


  if (Input.rte.mc.passback3D && Input.rte.mc.absorption!=MCABS_NONE && 
      !(Input.rte.mc.Nx_cloud * Input.rte.mc.Ny_cloud * Input.rte.mc.Nz_cloud)) {
    fprintf (stderr, "*** Error, need to specify mc_cloud_grid if mc_sample_grid and\n");
    fprintf (stderr, "***        mc_absorption are defined!\n");


    status--;
  }


  if (Input.rte.mc.absorption==MCABS_EMISSION && Input.source != SRC_THERMAL) {
    fprintf (stderr, "*** Error: mc_emission does only make sense with \"source thermal\"\n");
    status--;
  }


  /* check if zout is sorted in ascending order */
  for (i=1; i<Input.atm.nzout; i++)
    if (Input.atm.zout[i] - Input.atm.zout[i-1] < 0) {
      fprintf (stderr, "*** zout not sorted in ascending order\n");
      status--;
    }


  /* Check wavelength ranges */
  if (Input.wl.wl_start<0 || Input.wl.wl_end<0) {
    fprintf (stderr, "*** Error, need to define a wavelength range with 'wvn lower upper'!\n");
    status--;
  }

  if ( Input.spline && !(strlen(Input.filename[FN_SPLINE])>0) ) {
    if ( Input.wl.wl_start > Input.spline_lambda_0 )  {
      fprintf (stderr, "*** Spline wavelength %f is smaller than wvn %f\n",
	       Input.spline_lambda_0, Input.wl.wl_start);
      status--;
    }
    if ( Input.wl.wl_end < Input.spline_lambda_1 )  {
      fprintf (stderr, "*** Spline wavelength %f is greater than wvn %f\n",
	       Input.spline_lambda_1, Input.wl.wl_end);
      status--;
    }
  }

  /* Check that slit_function file exists */
  if ( Input.convolve ) {
    if ( (f = fopen(Input.filename[FN_SLITFUNCTION], "r")) == NULL) {
      fprintf (stderr, "*** Can not read slit function file %s\n", Input.filename[FN_SLITFUNCTION]);
      status--;
    }
    else 
      fclose(f);
  }

  
  /* Cloud checks */
  if ( Input.cloudcover != -999.0 ) {
    if (Input.cloudcover < 0.0 || Input.cloudcover > 1.0) {
      fprintf (stderr, "*** wc_cloudcover not between 0 and 1: %f\n", Input.cloudcover);
      status--;
    }
  }

  /* ck and libnetcdf */ 
#if HAVE_LIBNETCDF
#else
  if (Input.ck_scheme == CK_KATO) {
    fprintf (stderr, "   **********************************************************************\n");
    fprintf (stderr, "   * You have built uvspec without libnetcdf and hence may not          *\n");
    fprintf (stderr, "   * use the 'correlated_k kato' option. Please get netcdf and rebuild. *\n");
    fprintf (stderr, "   **********************************************************************\n");
    status--;
  }
#endif


  /* The MYSTIC solver */ 
#if HAVE_MYSTIC
#else
  if (Input.rte.solver == MONTECARLO) {
    fprintf (stderr, "   ********************************************************************\n");
    fprintf (stderr, "   * You have built uvspec without the three-dimensional MYSTIC       *\n");
    fprintf (stderr, "   * module. MYSTIC is currently not released to the public. If you   *\n");
    fprintf (stderr, "   * have an interest in three-dimensional calculations and if you    *\n");
    fprintf (stderr, "   * have a reasonable project in mind, please contact                *\n");
    fprintf (stderr, "   * bernhard.mayer@dlr.de.                                           *\n");
    fprintf (stderr, "   ********************************************************************\n");
    status--;
  }
#endif


  /* Fu and Liou */ 
#if HAVE_FULIOU
#else
  if (Input.ck_scheme == CK_FU) {
    fprintf (stderr, "   **********************************************************************\n");
    fprintf (stderr, "   * You have built uvspec without support for the Fu and Liou          *\n");
    fprintf (stderr, "   * parameterization. If you are interested, please contact the        *\n");
    fprintf (stderr, "   * libRadtran authors.                                                *\n");
    fprintf (stderr, "   **********************************************************************\n");
    status--;
  }
#endif


  /* Fu and Liou at present only works with water clouds defined per layer */
  if (Input.ck_scheme == CK_FU && strlen(Input.filename[FN_WCLOUD])>0 && !Input.wc.layer) {
    fprintf (stderr, "*** The Fu and Liou parameterization only works with cloud properties\n");
    fprintf (stderr, "*** defined per layer. Please add 'wc_layer' to the input file and   \n");
    fprintf (stderr, "*** check the documentation for its meaning!                         \n");

    status--;
  }

  /* Fu and Liou at present only works with ice clouds defined per layer */
  if (Input.ck_scheme == CK_FU && strlen(Input.filename[FN_ICLOUD])>0 && !Input.ic.layer) {
    fprintf (stderr, "*** The Fu and Liou parameterization only works with cloud properties\n");
    fprintf (stderr, "*** defined per layer. Please add 'ic_layer' to the input file and   \n");
    fprintf (stderr, "*** check the documentation for its meaning!                         \n");

    status--;
  }

  /* Ice clouds at present only work with Fu and Liou  */
  if ((Input.ck_scheme != CK_FU && strlen(Input.filename[FN_ICLOUD_PROP])==0) && strlen(Input.filename[FN_ICLOUD])>0) {
    fprintf (stderr, "*** The Fu and Liou ice cloud parameterization has been created for \n");
    fprintf (stderr, "*** the Fu and Liou 'correlated_k FU' bands; for other wavelengths  \n");
    fprintf (stderr, "*** grids please use 'ic_properties mie' or provide your own        \n");
    fprintf (stderr, "*** optical properties!                                             \n");

    status--;
  }

  
  return status;
}

